##########################################################################
#
# FILE:
#       updxfr.pm
# 
# DESCRIPTION: 
#       File transfer routines for upd
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       dd-mmm-yyyy, NN, first
#
##########################################################################
package updxfr;

use strict;
use Exporter();
use File::Path qw(mkpath);


use vars qw( @ISA @EXPORT);
@ISA = qw(Exporter);

# export this !
@EXPORT = qw (
      updxfr_ls
      updxfr_file
      updxfr_dir
      updxfr_put_unique
);

##########################################################################
# Imported Modules
use English;
use FileHandle;
use HTTP::Request::Common qw(POST);
use LWP::Simple qw($ua get getstore);
use upderr;
use updgbl;
use localftp;

# dont verify hostnames etc.
$ua->ssl_opts( verify_hostname => 0 ,SSL_verify_mode => 0x00);

##########################################################################
# Exported variables
#

##########################################################################
# Global and local variables
#
# use vars qw();

# my g_<global variables at file scope> = '' # initialization

##########################################################################
# Exported functions
#

#-------------------------------------------------------------------------
# <return> = updxfr_ls( node, dir , fd )
#
# <description>
#
# Input : nodename "node", directory path "dir"
# Output: filehandle to list upon "fd"
# Return: <return>
#
sub updxfr_ls
{
    my( $node,			# node name to do list upon
        $dir,			# directory to chdir to/ls
        $fd,			# file descriptor to put list upon
	$prop_info		# proprietary info, if any
    ) = @_;			

    &upderr_log( 4 , "updxfr_ls($node,$dir,...\n");

    my $url = "https://${node}/cgi-bin/updxfr.cgi?operation=ls&dir=${dir}";
    my $res = get($url);

    if (!defined($fd)) {
        print $res;
    } else {
        print {$fd} $res;
    }
}

sub updxfr_file
{
    my( $node,			# distribution node name to transfer from
        $dir,			# directory to chdir to there
        $dist, 			# file name on distribution node
        $local, 		# file name on local filesystem
	$prop_info		# proprietary login info
    ) = @_;			

    my ($local_dir, $osize, $size, $retry, $res, $url, $msg);
    &upderr_log( 4 , "updxfr_file($node,$dir,$dist,$local\n");

    # make directory for file if needed.

    ($local_dir) = ($local =~ m|(.*)/|o );

    if (defined($local_dir) && $local_dir ne "") {

	if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
	    upderr_log (1,"I would make directory $local_dir\n");
	} else {    
	  if (! -d $local_dir ) {
	    if (! eval {mkpath($local_dir, 0, 0777);} ) {
		upderr_log(0,"Unable to make directory $local_dir: $OS_ERROR\n");
		return 0;
	    }
	  }
	}
    }

    # initially we think the file is empty

    $osize = $size = 0;

    # I can no longer think why we wouldn't want to retry a transfer...
    # but we'll leave a variable here in case I do -- mengel
    $retry = 1;

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (1,"I would fetch $dist as $local from $node now\n");
	$res = 1;
    } else {
        # clean up node and dir...
        $node =~ s{:[0-9]*}{};
        $dir =~ s{^/ftp/}{/};

        $url = "https://${node}/cgi-bin/updxfr.cgi?operation=download&dir=${dir}/${dist}";
        $url =~ s{\+}{%2b}go;

        upderr_log(4, "fetching '$url'\n");

        $res = getstore($url, $local) == 200;

	unless (-f $local) {
	  $res = 0;
	  $msg = "unknown failure -- file did't exist after attempt";
	  upderr_log (1, "$local did not transfer\n"); 
	}
	if ($res) {
	    upderr_syslog( 'notice' , 
		"successful transfer ftp://${node}${dir}/${dist} -> $local\n" );
	    upderr_log(1, "informational: transferred ${dist}\n\tfrom ${node}:${dir} to\n\t${local}");
	} else {
	    upderr_syslog( 'warning' , 
			"failed transfer ftp://${node}${dir}/${dist} -> $local\n" );
	    upderr_syslog( 'warning' , "reason: $msg\n");
	    upderr_log(0, "error: can't transfer ${dir}/${dist}\n\tfrom ${node} to\n\t${local}");
	}
    }
    upderr_log( 4 , "returns $res\n");
    return $res;
}

sub updxfr_put_unique
{
    my( $node,			# distribution node name to transfer from
        $dir,			# directory to chdir to there
        $dist, 			# ref to return file name on distribution node
        $local, 		# file name on local filesystem
	$prop_info		# proprietary login info
    ) = @_;			
    my $res = 1;

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (1,"I would upload $local on $node now\n");
	$res = 1;
    } else {
        my ($req, $localbase);
        upderr_log("trying to put $local ");
        $req = $ua->request(POST "https://${node}/cgi-bin/updxfr.cgi",
            Content_Type => 'form-data',
            Content => [
                operation => 'upload',
                file => [$local],
            ]
        );
        $localbase = $local;
        $localbase =~ s{.*/}{};
        $$dist = ${localbase};
    }
    &upderr_log( 4 , "returns $res\n");
    return $res;
}

%updxfr::can_pipe_into = ( 
     'tar' => 1, 
     'tar.gz' => 1, 
     'zip' => 0, 
     'exe' => 0, 
     'shar' => 0 
);

# check for gnu tar as gtar...

my $tar;
if ( 0 == system("sh -c 'gtar --version' >/dev/null 2>&1") ) {
   $tar = "gtar";
} else {
   $tar = "tar";
}

%updxfr::unpack_command = ( 
    'tar'    => "${tar} xBf %s", 
    'tar.gz' => "gunzip -c %s | ${tar} xBf -", 
    'zip'    => 'unzip %s',
    'exe'    => '%s',
    'shar'   => 'sh %s'
);

%updxfr::verbose_unpack_command = ( 
    'tar'    => "${tar} xvBf %s", 
    'tar.gz' => "gunzip -c %s | ${tar} xvBf -", 
    'zip'    => 'unzip %s',
    'exe'    => '%s',
    'shar'   => 'sh -x %s'
);

sub updxfr_dir
{
    my( $node,
        $dir,
        $dist, 
        $local, 
	$prop_info,		# proprietary login info
	$arch_type		# archive type (tar, zip, etc.)
    ) = @_;			# arguments, as above
    my( $ftp,			# Net::FTP server handle
	$out,			# output filehandle
	$res,			# return code for this routine
	$intermediate,		# use an intermediate file
	$unpack_cmd,		# command to unwind archive
	$msg			# error message string
    );

    $res = 0;

    &upderr_log( 4 , "updxfr_dir($node,$dir,$dist,$local\n");
    &upderr_log( 4 , "            $prop_info, $arch_type)\n");

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (1,"I would make directory $local\n");
    } else {    
      if ( ! -d $local ) {
	if ( !eval {mkpath($local,0,0777);} ) {
             &upderr_log(0 , "unable to make directory '$local': $OS_ERROR\n");
	     return 0;
	}
      }
    }

    if ( !defined($arch_type) || $arch_type eq "" ) {
	$arch_type = "tar";
    }
    upderr_log (4,"arch_type is $arch_type\n");

    # use an intermediate file if we were told to, or if
    # the archive type can't be piped into

    $intermediate = ($g_flags{opt_I} || !($updxfr::can_pipe_into{$arch_type}));

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (1,"I would fetch directory $dist from ftp://$node/$dir as $local now\n"); 
	$res = 1;
    } else {
        if (! chdir($local)) {
           upderr_log(0,"Error: Could not chdir to $local\n");
           return 0;
        }
       
        &upderr_log(4,"Using intermediate file\n");
        # use an intermediate file to support retries...

        # fetch it to destination.tar
        $res = updxfr_file($node,	$dir,
                            $dist . "." . $arch_type, 
                            $local . "." . $arch_type);

        # blow it into our tar to unwind it.
        # we could catch unwind errors on the copy() or the close()

        if (upderr_getverbose() > 2) {
            $unpack_cmd = 
                sprintf($updxfr::verbose_unpack_command{$arch_type},
                        "$local.$arch_type" );
        } else {
            $unpack_cmd = 
                sprintf($updxfr::unpack_command{$arch_type},
                        "$local.$arch_type" );
        }

        upderr_log(3,"updxfr_dir: running unpack: $unpack_cmd");

        if (0 != system("$unpack_cmd")) {
            $res = 0;
            $msg = $!;
            &upderr_log(0,"Error: unwind failed: $msg\n");
        }

        unlink("$local.$arch_type");
        
    }
    if ($res) {
        upderr_syslog( 'notice' , 
            "successful unwind ftp://${node}${dir}/${dist} -> $local\n");
            upderr_log(1, "informational: transferred ${dir}/${dist}\n\tfrom ${node} to\n\t${local}");
    } else {
        upderr_syslog( 'warning' , 
            "failed unwind ftp://${node}${dir}/${dist} -> $local\n");
        upderr_syslog( 'warning' , "reason: $msg\n");
        upderr_log(0, "error: can't transfer ${dir}/${dist}\n\t from ${node} to \n\t${local}");
    }
    &upderr_log( 4 , "returns $res\n");
    return $res;
}


# If standalone, run tests
if ( $0 =~ /updxfr.pm$/ ) {

   upderr_setverbose(9);
   my @prop_info;
   my $node;

   my $errstr;

   if ( $#ARGV == -1 ) {

       print "default regression test:\n";

       $node = "ftp.fnal.gov";
       #$node = "fermicloud260.fnal.gov";

       print "listing /etc on $node\n";

       print("ls of /etc:\n");
       &updxfr_ls($node,'/etc', \*STDOUT, $prop_info[0]);
       upderr_get(\$errstr); print STDERR $errstr;
       print("-----\n");

       print("ls of /KITS:\n");
       &updxfr_ls($node,'/KITS', \*STDOUT, $prop_info[1]);
       upderr_get(\$errstr); print STDERR $errstr;
       print("-----\n");

       #print "fetching .proplogfile from ftp.fnal.gov\n";
       #&updxfr_file($node,'/KITS','README.kits','testfile','');
       #upderr_get(\$errstr); print STDERR $errstr;

       #print "fetching bogus file from ftp.fnal.gov\n";
       #&updxfr_file($node,'/KITS','nosuchfile','testfile2',$prop_info[0]);
       #upderr_get(\$errstr); print STDERR $errstr;

       print "putting testfile1 file to incoming\n";
       &updxfr_put_unique($node,'/incoming','testfile2','updxfr.pm','');
       print "ls of /incoming\n";
       &updxfr_ls($node,'/incoming', \*STDOUT, $prop_info[1]);

       exit(0);

       print "fetching and unwinding tarfile from ftp.fnal.gov\n";
       &updxfr_dir($node,'/KITS/GENERIC_UNIX/ups_compile/v1_3','ups_compilev1_3', 'testdir',$prop_info[0]);
       upderr_get(\$errstr); print STDERR $errstr;

       print "Done.\n"
    } elsif ( $ARGV[0] eq "-f" ) {
       shift(@ARGV);
       print "fetching $ARGV[0]/$ARGV[1]  as $ARGV[1]\n";
       &updxfr_file($node, $ARGV[0], $ARGV[1], $ARGV[2],$prop_info[0]);
       print "Done.\n";
    } elsif ( $ARGV[0] eq "-i" ) {
        my $host = "ftp" ;
	my $dir = "/";
	my $dist = "README";
	my $local = "/tmp/junk";

       print "> "; 
       while (<STDIN>) {
	   if (/local *= *(.*)/) { $local = $1; }
	   elsif (  /dir *= *(.*)/) {   $dir = $1; }
	   elsif (  /cd *(.*)/)     {   $dir .= "/" . $1; }
	   elsif ( /dist *= *(.*)/) {  $dist = $1; }
	   elsif ( /host *= *(.*)/) {  $host = $1; }
	   elsif ( /^xf$|file/ ) { updxfr_file($host,$dir,$dist,$local,$prop_info[0]);}
	   elsif ( /^xd$|dir/ ) { updxfr_dir($host,$dir,$dist,$local,$prop_info[0]);}
	   elsif ( /^xp$|put_unique/ ) { updxfr_put_unique($host,$dir,\$dist,$local,$prop_info[0]);}
	   elsif ( /^xl$|ls/ ) { updxfr_ls($host,$dir, $main::STDOUT, $prop_info[0]); }
	   elsif ( /^quit|exit/) { exit 0; }
	   else { print "unkown command: $_\n";}
	   upderr_get(); upderr_clear();

	   print "local: $local\n";
	   print "host: $host\n";
	   print "dir: $dir\n";
	   print "dist: $dist\n";
           print "> "; 
       }
    }
}
1;

__END__

=head1 NAME

updxfr - blah blah

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION

<description>

=head1 BUGS

<bugs>

=cut
